/*
 * Moblin-Web-Browser: The web browser for Moblin
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 */

const MwbIPagesObserver
  = Components.interfaces.MwbIPagesObserver;
const nsISupports = Components.interfaces.nsISupports;
const nsISupportsWeakReference = Components.interfaces.nsISupportsWeakReference;

var pagesService;
var searchStr = "";

function PinnedPagesObserver()
{
}

PinnedPagesObserver.prototype =
{
  favoritesReceived : function(favorites, nFavorites)
  {
  },

  linkVisited : function(uri, visit_time, is_pinned)
  {
    if (is_pinned)
      rebuildPinnedPagesList();
  },

  pinnedPage : function(title, uri, visit_time)
  {
    rebuildPinnedPagesList();
  },

  unpinnedPage : function(uri)
  {
    rebuildPinnedPagesList();
  },

  QueryInterface : function(aIID)
  {
    if (!aIID.equals(MwbIPagesObserver)
        && !aIID.equals(nsISupports)
        && !aIID.equals(nsISupportsWeakReference))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};

function searchInString(haystack, needle)
{
  var i;

  if (needle.length <= haystack.length)
    for (i = 0; i <= haystack.length - needle.length; i++)
      {
        var j;
        for (j = 0; j < needle.length; j++)
          if (haystack.charAt(i + j) != needle.charAt(j))
            break;
        if (j >= needle.length)
          return i;
      }

  return -1;
}

function rebuildPinnedPagesList()
{
  var pinnedPagesDiv = document.getElementById("pinned_pages_div");
  var lowerSearchStr;
  var pageCount = 0;
  var i;

  // Remove all existing pinned pages
  pinnedPagesDiv.innerHTML = "";

  if (searchStr && searchStr.match(/\S/))
    lowerSearchStr = searchStr.trim().toLowerCase();

  for (i = 0; i < pagesService.numPinnedPages; i++)
    {
      var page = pagesService.getPinnedPageByTitle(i);
      var titleMatch;

      if (!lowerSearchStr ||
          (titleMatch = searchInString(page.title.toLowerCase(),
                                       lowerSearchStr)) != -1)
        {
          var div = document.createElement("div");
          div.setAttribute("class", "pinned_page_div");

          var a = document.createElement("a");
          a.setAttribute("href", page.uri);

          var thumbnailDiv = document.createElement("div");
          thumbnailDiv.setAttribute("class", "pinned_page_thumbnail");

          // Create elements for the pin button. The outer div is just
          // used to make the position for the inner element relative to
          // the top of the thumbnail
          var pinOuter = document.createElement('div');
          pinOuter.setAttribute('class', 'pin-icon-outer');
          var pinInner = document.createElement('a');
          pinInner.setAttribute('class', 'pin-icon-inner');
          pinInner.setAttribute('href', 'javascript:');
          pinOuter.appendChild(pinInner);
          thumbnailDiv.appendChild(pinOuter);

          // Create a closure with the url and use it as a callback
          // to unpin the page when the link is clicked
          let (uri = page.uri) {
            pinInner.addEventListener("click",
                                      function(e) {
                                        pagesService.unpinPage(uri);
                                        pagesService.getPinnedPages();
                                        e.preventDefault();
                                        e.stopPropagation();
                                      },
                                      false);
          }

          var img = document.createElement("img");
          img.setAttribute("class", "thumbnail");
          img.setAttribute("src", pagesService.getThumbnailURL(page.uri).spec);
          thumbnailDiv.appendChild(img);

          a.appendChild(thumbnailDiv);

          img = document.createElement("img");
          img.setAttribute("class", "pinned_page_favicon");
          img.setAttribute("src", "mwb-favicon://" + page.uri);
          a.appendChild(img);

          var span = document.createElement("span");
          span.setAttribute("class", "pinned_page_title");
          if (titleMatch != undefined)
            {
              var title = page.title;
              var preMatch =
                document.createTextNode(title.slice(0, titleMatch));
              span.appendChild(preMatch);
              var match =
                document.createTextNode(title.slice(titleMatch,
                                                    lowerSearchStr.length +
                                                    titleMatch));
              var matchSpan = document.createElement("span");
              matchSpan.setAttribute("class", "search-match");
              matchSpan.appendChild(match);
              span.appendChild(matchSpan);
              var postMatch =
                document.createTextNode(title.slice(titleMatch +
                                                    lowerSearchStr.length));
              span.appendChild(postMatch);
            }
          else
            span.appendChild(document.createTextNode(page.title));
          a.appendChild(span);

          var timeDiv = document.createElement("div");
          timeDiv.setAttribute("class", "pinned_page_time");
          var timeStr = pagesService.formatTime(page.visitTime);
          timeDiv.appendChild(document.createTextNode(timeStr));
          a.appendChild(timeDiv);

          div.appendChild(a);

          pinnedPagesDiv.appendChild(div);

          pageCount++;
        }
    }

  if (pageCount <= 0)
    {
      var div = document.createElement("div");
      var note =
        lowerSearchStr ?
        _("No matching pages were found") :
        _("Push the pin icon in the location bar above to " +
          "add pages to this list.");
      div.appendChild(document.createTextNode(note));
      div.setAttribute("class", "pinned_pages_note");
      pinnedPagesDiv.appendChild(div);
    }
}

function updateSearch()
{
  searchStr = document.getElementById("search-entry").value;
  rebuildPinnedPagesList();
}

function searchEntryKeyDownCb(event)
{
  // Update the search when enter is pressed
  if (event.keyCode == 13)
    updateSearch();
}

var observer;

// Run the initialisation once the page has finished loading
document.addEventListener("DOMContentLoaded", function(event)
{
  pagesService = (Components.classes["@moblin.org/mwb-pages-service;1"].
                  getService(Components.interfaces.MwbIPagesService));

  observer = new PinnedPagesObserver();

  // Add our observer for when the favorites become available
  pagesService.addPagesObserver(observer);

  // Start fetching the pinned pages immediately
  pagesService.getPinnedPages();
  rebuildPinnedPagesList();

  var searchButton = document.getElementById("search-button");
  searchButton.addEventListener("click", updateSearch, false);
  var searchEntry = document.getElementById("search-entry");
  searchEntry.addEventListener("keydown", searchEntryKeyDownCb, false);

  // Unregister the load event so we don't initialize more than once
  event.currentTarget.removeEventListener("DOMContentLoaded",
                                          arguments.callee,
                                          false);
}, false);
